#include <linux/genetlink.h>
#include <linux/nldlm.h>
#include <getopt.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>

#include "libnldlm.h"
#include "ynl/ynl.h"

struct nldlmw_ctx {
	struct ynl_sock *ys;
	void (*ls_event_cb)(__u8 cmd, const struct nldlm_get_ls_rsp *ls,
			    void *data);
	void *ls_event_data;
};

int nldlmw_dump_lockspaces(struct nldlmw_ctx *ctx,
			   void (*cb)(const struct nldlm_get_ls_rsp *ls))
{
	struct nldlm_get_ls_list *lss;

	/* error handling? */
	lss = nldlm_get_ls_dump(ctx->ys);
	if (!lss)
		return -1;

	while (!ynl_dump_empty(lss)) {
		cb(&lss->obj);
		lss = lss->next;
	}

	nldlm_get_ls_list_free(lss);
	return 0;
}

int nldlmw_dump_members(struct nldlmw_ctx *ctx, const char *lsname,
			void (*cb)(const struct nldlm_get_ls_member_rsp *mb))
{
	struct nldlm_get_ls_member_req_dump *req;
	struct nldlm_get_ls_member_list *ls_mbs;

	req = nldlm_get_ls_member_req_dump_alloc();
	nldlm_get_ls_member_req_dump_set_ls_name(req, lsname);

	/* error handling? */
	ls_mbs = nldlm_get_ls_member_dump(ctx->ys, req);
	nldlm_get_ls_member_req_dump_free(req);
	if (!ls_mbs)
		return -1;

	while (!ynl_dump_empty(ls_mbs)) {
		cb(&ls_mbs->obj);
		ls_mbs = ls_mbs->next;
	}

	nldlm_get_ls_member_list_free(ls_mbs);
	return 0;
}

int nldlmw_ls_add_member(struct nldlmw_ctx *ctx, const char *lsname,
			 unsigned int nodeid)
{
	struct nldlm_ls_add_member_req *req;
	int rv;

	req = nldlm_ls_add_member_req_alloc();
	nldlm_ls_add_member_req_set_ls_name(req, lsname);
	nldlm_ls_add_member_req_set_nodeid(req, htole32(nodeid));

	rv = nldlm_ls_add_member(ctx->ys, req);

	nldlm_ls_add_member_req_free(req);
	return rv;
}

int nldlmw_ls_del_member(struct nldlmw_ctx *ctx, const char *lsname,
			 unsigned int nodeid)
{
	struct nldlm_ls_del_member_req *req;
	int rv;

	req = nldlm_ls_del_member_req_alloc();
	nldlm_ls_del_member_req_set_ls_name(req, lsname);
	nldlm_ls_del_member_req_set_nodeid(req, htole32(nodeid));

	rv = nldlm_ls_del_member(ctx->ys, req);

	nldlm_ls_del_member_req_free(req);
	return rv;
}

static int nldlmw_ls_event_done(struct nldlmw_ctx *ctx, const char *lsname,
				unsigned int ls_global_id,
				bool has_global,
				enum nldlm_ls_event_result result)
{
	struct nldlm_ls_event_done_req *req;
	int rv;

	req = nldlm_ls_event_done_req_alloc();
	nldlm_ls_event_done_req_set_ls_name(req, lsname);
	if (has_global)
		nldlm_ls_event_done_req_set_ls_global_id(req, htole32(ls_global_id));
	nldlm_ls_event_done_req_set_result(req, result);

	rv = nldlm_ls_event_done(ctx->ys, req);

	nldlm_ls_event_done_req_free(req);
	return rv;
}

int nldlmw_ls_event_done_success(struct nldlmw_ctx *ctx, const char *lsname,
				 unsigned int ls_global_id)
{
	return nldlmw_ls_event_done(ctx, lsname, ls_global_id, true,
				    NLDLM_LS_EVENT_RESULT_SUCCESS);
}

int nldlmw_ls_event_done_failure(struct nldlmw_ctx *ctx, const char *lsname)
{
	return nldlmw_ls_event_done(ctx, lsname, -1 /* not used */, false,
				    NLDLM_LS_EVENT_RESULT_FAILURE);
}

int nldlmw_ls_control(struct nldlmw_ctx *ctx, const char *lsname,
		      enum nldlm_ls_ctrl_action action)
{
	struct nldlm_ls_ctrl_req *req;
	int rv;

	req = nldlm_ls_ctrl_req_alloc();
	nldlm_ls_ctrl_req_set_ls_name(req, lsname);
	nldlm_ls_ctrl_req_set_action(req, action);

	rv = nldlm_ls_ctrl(ctx->ys, req);

	nldlm_ls_ctrl_req_free(req);
	return rv;
}

/* TODO why they are not generated? */
static inline struct nldlm_addr *nldlmw_addr_alloc(unsigned int n_addrs)
{
	return calloc(n_addrs, sizeof(struct nldlm_addr));
}

static inline void
nldlmw_addr_set_family(struct nldlm_addr *addr, __u16 family)
{
	addr->_present.family = 1;
	addr->family = family;
}

static inline void
nldlmw_addr_set_addr4(struct nldlm_addr *addr,
		      __u32 addr4 /* big-endian */)
{
	addr->_present.addr4 = 1;
	addr->addr4 = addr4;
}

/* free is done by nldlm_add_node_req_free() */

int nldlmw_new_node(struct nldlmw_ctx *ctx, unsigned int nodeid,
		    const struct sockaddr_storage *addr)
{
	struct nldlm_add_node_req *req;
	struct nldlm_addr *nladdr;
	int rv;

	req = nldlm_add_node_req_alloc();
	nldlm_add_node_req_set_id(req, htole32(nodeid));
	nladdr = nldlmw_addr_alloc(1);
	nldlmw_addr_set_family(&nladdr[0], AF_INET);
	nldlmw_addr_set_addr4(&nladdr[0], ((struct sockaddr_in *)addr)->sin_addr.s_addr);
	__nldlm_add_node_req_set_addrs(req, nladdr, 1);

	rv = nldlm_add_node(ctx->ys, req);

	nldlm_add_node_req_free(req);
	return rv;
}

int nldlmw_dump_nodes(struct nldlmw_ctx *ctx,
		      void (*cb)(const struct nldlm_get_node_rsp *nd))
{
	struct nldlm_get_node_list *nds;

	/* error handling? */
	nds = nldlm_get_node_dump(ctx->ys);
	if (!nds)
		return -1;

	while (!ynl_dump_empty(nds)) {
		cb(&nds->obj);
		nds = nds->next;
	}

	nldlm_get_node_list_free(nds);
	return 0;
}

int nldlmw_set_cluster_name(struct nldlmw_ctx *ctx, const char *cluster_name)
{
	struct nldlm_set_cluster_name_req *req;
	int rv;

	req = nldlm_set_cluster_name_req_alloc();
	nldlm_set_cluster_name_req_set_cluster_name(req, cluster_name);

	rv = nldlm_set_cluster_name(ctx->ys, req);

	nldlm_set_cluster_name_req_free(req);
	return rv;
}

int nldlmw_set_recover_callbacks(struct nldlmw_ctx *ctx, int on)
{
	struct nldlm_set_recover_callbacks_req *req;
	int rv;

	req = nldlm_set_recover_callbacks_req_alloc();
	if (on)
		nldlm_set_recover_callbacks_req_set_recover_callbacks(req);

	rv = nldlm_set_recover_callbacks(ctx->ys, req);

	nldlm_set_recover_callbacks_req_free(req);
	return rv;
}

int nldlmw_set_our_nodeid(struct nldlmw_ctx *ctx, unsigned int nodeid)
{
	struct nldlm_set_our_nodeid_req *req;
	int rv;

	req = nldlm_set_our_nodeid_req_alloc();
	nldlm_set_our_nodeid_req_set_our_nodeid(req, htole32(nodeid));

	rv = nldlm_set_our_nodeid(ctx->ys, req);

	nldlm_set_our_nodeid_req_free(req);
	return rv;
}

int nldlmw_get_config(struct nldlmw_ctx *ctx,
		      void (*cb)(const struct nldlm_get_cfg_rsp *cfg))
{
	struct nldlm_get_cfg_rsp *rsp;

	rsp = nldlm_get_cfg(ctx->ys);
	if (!rsp)
		return -1;

	cb(rsp);

	nldlm_get_cfg_rsp_free(rsp);
	return 0;
}

int nldlmw_subscribe_events(struct nldlmw_ctx *ctx, const char *group,
			    void (*cb)(__u8 cmd, const struct nldlm_get_ls_rsp *ls, void *data),
			    void *data)
{
	int rv;

	rv = ynl_subscribe(ctx->ys, group);
	if (!rv) {
		ctx->ls_event_cb = cb;
		ctx->ls_event_data = data;
	}

	return rv;
}

int nldlmw_process(struct nldlmw_ctx *ctx)
{
	struct nldlm_get_ls_ntf *ntf_ls;
	struct ynl_ntf_base_type *ntf;
	int rv;

	rv = ynl_ntf_check(ctx->ys);
	if (rv)
		return -1;

	ntf = ynl_ntf_dequeue(ctx->ys);
	while (ntf) {
		switch (ntf->cmd) {
		case NLDLM_CMD_NTF_NEW_LS:
			/* fallthrough */
		case NLDLM_CMD_NTF_RELEASE_LS:
			ntf_ls = (struct nldlm_get_ls_ntf *)ntf;
			ctx->ls_event_cb(ntf_ls->cmd, &ntf_ls->obj,
					 ctx->ls_event_data);
			break;
		default:
			/* ignore */
			break;
		}

		ynl_ntf_free(ntf);
		ntf = ynl_ntf_dequeue(ctx->ys);
	}

	return 0;
}

int nldlmw_get_fd(const struct nldlmw_ctx *ctx)
{
	return ynl_socket_get_fd(ctx->ys);
}

int nldlmw_init(struct nldlmw_ctx **ctx)
{
	struct nldlmw_ctx *c;
	struct ynl_error err;

	*ctx = NULL;

	c = calloc(1, sizeof(*c));
	if (!c)
		return -ENOMEM;

	c->ys = ynl_sock_create(&ynl_nldlm_family, &err);
	if (!c->ys)
		return -1;

	*ctx = c;
	return 0;
}

void nldlmw_exit(struct nldlmw_ctx *c)
{
	ynl_sock_destroy(c->ys);
	free(c);
}
