#include <linux/nldlm.h>
#include <sys/epoll.h>
#include <inttypes.h>
#include <stdbool.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "libnldlm.h"

static char usage_str[] = {
"\n"
"  -h, --help              Show this help screen.\n"
"  -v, --version           Print the version and quit.\n"
};

static void usage(FILE *o, const char *pname)
{
	fprintf(o, "usage: %s %s\n", pname, usage_str);
}

static void print_node(const struct nldlm_get_node_rsp *nd)
{
	if (!nd->_present.id) {
		printf("invalid\n");
	} else {
		printf("%" PRIu32 ":\n", le32toh(nd->id));
	}

	if (!nd->_present.mark) {
		printf("invalid\n");
	} else {
		printf("\tmark: %" PRIu32 "\n", nd->mark);
	}

	if (!nd->n_addrs) {
		printf("invalid\n");
	} else {
		printf("\taddrs:\n");
	}
}

static void print_lockspace(const struct nldlm_get_ls_rsp *ls)
{
	printf("\"%s\":\n", ls->name);
}

static void print_member(const struct nldlm_get_ls_member_rsp *mb)
{
	if (!mb->_present.nodeid) {
		printf("invalid\n");
	} else {
		printf("%" PRIu32 ":\n", le32toh(mb->nodeid));
	}

	if (!mb->_present.weight) {
		printf("invalid\n");
	} else {
		printf("\tweight: %" PRIu32 "\n", mb->weight);
	}
}

static int do_dump(struct nldlmw_ctx *nlctx, int argc, char *argv[])
{
	const char *lsname;
	const char *opt_s;

	opt_s = argv[optind++];
	if (!strcmp(opt_s, "ls"))
		return nldlmw_dump_lockspaces(nlctx, print_lockspace);
	else if (!strcmp(opt_s, "node"))
		return nldlmw_dump_nodes(nlctx, print_node);

	/* requires lockspace name */
	if (argc <= optind) {
		usage(stderr, argv[0]);
		return -1;
	}
	lsname = argv[optind++];

	if (!strcmp(opt_s, "member"))
		return nldlmw_dump_members(nlctx, lsname, print_member);

	usage(stderr, argv[0]);
	return -1;
}

static void print_config(const struct nldlm_get_cfg_rsp *cfg)
{
	printf("our node: %u\n", le32toh(cfg->our_nodeid));
	printf("cluster_name: \"%s\"\n", cfg->cluster_name);
	printf("protocol: \"%s\"\n", nldlm_protocol_str(cfg->protocol));
	printf("port: %" PRIu16 "\n", ntohs(cfg->port));
	printf("log level: %" PRIu32 "\n", cfg->log_level);
	printf("recover timeout: %" PRIu32 "\n", cfg->recover_timeout);
	printf("inactive timeout: %" PRIu32 "\n", cfg->inactive_timeout);
	printf("default mark: %" PRIu32 "\n", cfg->default_mark);
	printf("recover_callbacks: %u\n", cfg->_present.recover_callbacks);
}

static int do_get(struct nldlmw_ctx *nlctx, int argc, char *argv[])
{
	const char *opt_s;

	opt_s = argv[optind++];
	if (!strcmp(opt_s, "cfg"))
		return nldlmw_get_config(nlctx, print_config);

	usage(stderr, argv[0]);
	return -1;
}

static void print_ls_event(__u8 cmd, const struct nldlm_get_ls_rsp *ls, void *data)
{
	switch (cmd) {
	case NLDLM_CMD_NTF_NEW_LS:
		printf("lockspace \"%s\" new\n", ls->name);
		break;
	case NLDLM_CMD_NTF_RELEASE_LS:
		printf("lockspace \"%s\" release\n", ls->name);
		break;
	default:
		printf("lockspace \"%s\" unknown event\n", ls->name);
		break;
	}
}

static int do_monitor(struct nldlmw_ctx *nlctx, int argc, char *argv[])
{
	return nldlmw_subscribe_events(nlctx, "ls-event", print_ls_event, NULL);
}

#define MAX_EVENTS 10
struct epoll_event ev, events[MAX_EVENTS];
int nfds, epollfd;

int main(int argc, char *argv[])
{
	struct nldlmw_ctx *nlctx;
	const char *opt_s;
	bool loop = false;
	int opt, rv;

	while ((opt = getopt(argc, argv, "vh")) != -1) {
		switch (opt) {
		case 'v':
			break;
		case 'h':
			usage(stderr, argv[0]);
			exit(EXIT_SUCCESS);
		default:
			usage(stderr, argv[0]);
			exit(EXIT_FAILURE);
		}
	}

	if (argc <= optind) {
		usage(stderr, argv[0]);
		exit(EXIT_FAILURE);
	}

	rv = nldlmw_init(&nlctx);
	if (rv < 0)
		exit(EXIT_FAILURE);

	opt_s = argv[optind++];
	if (!strcmp(opt_s, "dump")) {
		rv = do_dump(nlctx, argc, argv);
	} else if (!strcmp(opt_s, "get")) {
		rv = do_get(nlctx, argc, argv);
	} else if (!strcmp(opt_s, "monitor")) {
		loop = true;
		rv = do_monitor(nlctx, argc, argv);
	} else {
		rv = -1;
	}

	if (rv < 0) {
		exit(EXIT_FAILURE);
	}

	epollfd = epoll_create1(0);
	if (epollfd == -1) {
		perror("epoll_create1");
		exit(EXIT_FAILURE);
	}

	ev.events = EPOLLIN;
	if (epoll_ctl(epollfd, EPOLL_CTL_ADD, nldlmw_get_fd(nlctx), &ev) == -1) {
		perror("epoll_ctl: listen_sock");
		exit(EXIT_FAILURE);
	}

	while (loop) {
		nfds = epoll_wait(epollfd, events, MAX_EVENTS, -1);
		if (nfds == -1) {
			perror("epoll_wait");
			exit(EXIT_FAILURE);
		}

		rv = nldlmw_process(nlctx);
		if (rv < 0 && errno != EAGAIN)
			loop = false;
	}

	nldlmw_exit(nlctx);

	if (rv < 0) {
		usage(stderr, argv[0]);
		exit(EXIT_FAILURE);
	}

	return EXIT_SUCCESS;
}

