#define _GNU_SOURCE

#include <sys/statvfs.h>
#include <linux/nldlm.h>
#include <sys/types.h>
#include <sys/mount.h>
#include <arpa/inet.h>
#include <sys/vfs.h>
#include <dirent.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <sched.h>
#include <ev.h>

#include <lua.h>
#include <lauxlib.h>
#include <lualib.h>

#include "libnldlm.h"
#include "helpers.h"
#include "list.h"

#define COALESCENCE_MEMBER_UPDATE_SECS 1.
#define NETNS_RUN_DIR "/var/run/netns"

struct dlm_op {
	struct node *n;
	uint32_t type;

	struct list_head list;
};

struct node {
	struct sockaddr_storage addr;
	struct nldlmw_ctx *nlctx;
	unsigned int id;
	ev_io w;
	int fd;

	struct list_head list;
	char nsname[0];
};

struct dlm_member {
	struct node *node;

	struct list_head list;
};

struct dlm_ls {
	char lsname[64];
	struct list_head members;
	struct list_head ops;
	unsigned int ls_global_id;

	struct list_head list;
};

static unsigned int global_id;
static struct ev_loop *loop;
static bool timer_done;
static ev_timer timer;

static LIST_HEAD(ls_list);
static LIST_HEAD(nodes_list);

static struct node *get_node_by_id(unsigned int id)
{
	struct node *n;

	if (!id)
		return NULL;

	list_for_each_entry(n, &nodes_list, list) {
		if (n->id == id)
			return n;
	}

	return NULL;
}

static struct node *get_node_by_nsname(const char *nsname)
{
	struct node *n;

	if (!nsname)
		return NULL;

	list_for_each_entry(n, &nodes_list, list) {
		if (!strcmp(n->nsname, nsname))
			return n;
	}

	return NULL;
}

static int new_node(const char *nsname, unsigned int id, struct node **ret)
{
	struct node *n;

	n = get_node_by_nsname(nsname);
	if (n)
		return -EEXIST;

	n = get_node_by_id(id);
	if (n)
		return -EEXIST;

	n = calloc(1, sizeof(*n) + strlen(nsname) + 1);
	if (!n) {
		perror("malloc");
		return -ENOMEM;
	}

	strcpy(n->nsname, nsname);
	n->id = id;
	*ret = n;
	return 0;
}

#if 0
static void del_node(struct nsid_cache *c)
{
	list_del(&c->list);
	free(c);
}
#endif

static int config_load(const char *filename)
{
	struct sockaddr_in *in_addr;
	const char *nsname, *addr;
	unsigned int nodeid;
	struct node *n;
	lua_State *L;
	int rv;

	L = luaL_newstate();
	if (!L)
		return -1;

	luaopen_base(L);
	luaopen_io(L);
	luaopen_string(L);
	luaopen_math(L);

	if (luaL_loadfile(L, filename) || lua_pcall(L, 0, 0, 0)) {
		fprintf(stderr, "cannot run configuration file: %s",
			lua_tostring(L, -1));
		return -1;
	}

	lua_getglobal(L, "nodes");
	if (!lua_istable(L, -1))
		return -1;

	lua_pushnil(L);
	while (lua_next(L, -2) != 0) {
		lua_getfield(L, -1, "nodeid");
		if (!lua_isnumber(L, -1))
			return -1;

		nodeid = lua_tonumber(L, -1);
		lua_pop(L, 1);

		lua_getfield(L, -1, "nsname");
		if (!lua_isstring(L, -1))
			return -1;

		nsname = lua_tostring(L, -1);
		lua_pop(L, 1);

		lua_getfield(L, -1, "addr");
		if (!lua_isstring(L, -1))
			return -1;

		addr = lua_tostring(L, -1);
		rv = new_node(nsname, nodeid, &n);
		if (rv < 0)
			return -1;

		in_addr = (struct sockaddr_in *)&n->addr;
		inet_pton(AF_INET, addr, &in_addr->sin_addr);
		printf("Configure nsname: %s id: %u addr: %s\n", nsname, nodeid, addr);
		list_add_tail(&n->list, &nodes_list);
		lua_pop(L, 1);

		lua_pop(L, 1);
	}
	lua_pop(L, 1);

	return 0;
}

static struct dlm_ls *add_or_get_ls(struct list_head *head, const char *lsname)
{
	struct dlm_ls *iter, *ls = NULL;

	if (!lsname)
		return NULL;

	list_for_each_entry(iter, head, list) {
		if (!strcmp(iter->lsname, lsname)) {
			ls = iter;
			break;
		}
	}

	if (!ls) {
		ls = calloc(1, sizeof(*ls));
		if (!ls)
			return NULL;

		INIT_LIST_HEAD(&ls->members);
		strcpy(ls->lsname, lsname);
		INIT_LIST_HEAD(&ls->ops);
		/* global id is just a counter */
		ls->ls_global_id = ++global_id;

		list_add_tail(&ls->list, head);
	}

	return ls;
}

static int ls_add_member(struct dlm_ls *ls, struct node *n)
{
	struct dlm_member *memb;

	if (!n)
		return -EINVAL;

	list_for_each_entry(memb, &ls->members, list) {
		if (memb->node->id == n->id)
			return -EEXIST;
	}

	memb = calloc(1, sizeof(*memb));
	if (!memb)
		return -ENOMEM;

	memb->node = n;
	list_add_tail(&memb->list, &ls->members);
	return 0;
}

static int ls_del_member(struct dlm_ls *ls, struct node *n)
{
	struct dlm_member *memb;

	if (!n)
		return -EINVAL;

	list_for_each_entry(memb, &ls->members, list) {
		if (memb->node->id == n->id) {
			list_del(&memb->list);
			free(memb);
			return 0;
		}
	}

	return -ENOENT;
}

static void timeout_cb(EV_P_ ev_timer *t, int revents)
{
	struct dlm_op *op, *op2, *safe;
	struct dlm_member *memb;
	struct dlm_ls *ls;
	int rv;

	list_for_each_entry(ls, &ls_list, list) {
		list_for_each_entry(op, &ls->ops, list) {
			switch (op->type) {
			case NLDLM_CMD_NTF_NEW_LS:
				list_for_each_entry(op2, &ls->ops, list) {
					nldlmw_ls_add_member(op->n->nlctx, ls->lsname, op2->n->id);
				}

				list_for_each_entry(memb, &ls->members, list) {
					nldlmw_ls_control(memb->node->nlctx, ls->lsname,
							  NLDLM_LS_CTRL_ACTION_STOP);
				}

				list_for_each_entry(memb, &ls->members, list) {
					/* all current to ourself */
					nldlmw_ls_add_member(op->n->nlctx, ls->lsname, memb->node->id);
					/* ourself to all current */
					nldlmw_ls_add_member(memb->node->nlctx, ls->lsname, op->n->id);
				}

				list_for_each_entry(memb, &ls->members, list) {
					nldlmw_ls_control(memb->node->nlctx, ls->lsname,
							 NLDLM_LS_CTRL_ACTION_START);
				}

				nldlmw_ls_control(op->n->nlctx, ls->lsname,
						 NLDLM_LS_CTRL_ACTION_START);
				break;
			case NLDLM_CMD_NTF_RELEASE_LS:
				list_for_each_entry(memb, &ls->members, list) {
					if (memb->node->id == op->n->id)
						continue;

					nldlmw_ls_control(memb->node->nlctx, ls->lsname,
							 NLDLM_LS_CTRL_ACTION_STOP);
				}

				list_for_each_entry(memb, &ls->members, list) {
					if (memb->node->id == op->n->id)
						continue;

					/* ourself to all current */
					nldlmw_ls_del_member(memb->node->nlctx, ls->lsname, op->n->id);
				}

				list_for_each_entry(memb, &ls->members, list) {
					if (memb->node->id == op->n->id)
						continue;

					nldlmw_ls_control(memb->node->nlctx, ls->lsname,
							 NLDLM_LS_CTRL_ACTION_START);
				}
				break;
			default:
				printf("BUG %u\n", __LINE__);
				goto out;
			}
		}
	}

	list_for_each_entry(ls, &ls_list, list) {
		list_for_each_entry(op, &ls->ops, list) {
			switch (op->type) {
			case NLDLM_CMD_NTF_NEW_LS:
				rv = ls_add_member(ls, op->n);
				if (rv < 0)
					printf("BUG %d %u\n", rv, __LINE__);
				break;
			case NLDLM_CMD_NTF_RELEASE_LS:
				rv = ls_del_member(ls, op->n);
				if (rv < 0)
					printf("BUG %d %u\n", rv, __LINE__);
				break;
			default:
				printf("BUG %u\n", __LINE__);
				goto out;
			}
		}
	}

out:
	list_for_each_entry(ls, &ls_list, list) {
		list_for_each_entry_safe(op, safe, &ls->ops, list) {
			list_del(&op->list);
			free(op);
		}
	}
	timer_done = false;
}

static void print_ls_event(__u8 cmd, const struct nldlm_get_ls_rsp *els, void *data)
{
	struct node *n = data;
	struct dlm_ls *ls;
	struct dlm_op *op;

	ls = add_or_get_ls(&ls_list, els->name);
	if (!ls)
		goto out;

	switch (cmd) {
        case NLDLM_CMD_NTF_NEW_LS:
		list_for_each_entry(op, &ls->ops, list) {
			if (op->n == n && op->type == NLDLM_CMD_NTF_RELEASE_LS) {
					list_del(&op->list);
					goto out;
			}
		}

		op = calloc(1, sizeof(*op));
		if (!op) {
			nldlmw_ls_event_done_failure(n->nlctx, els->name);
			return;
		}

		op->type = cmd;
		op->n = n;

		list_add_tail(&op->list, &ls->ops);
		break;
        case NLDLM_CMD_NTF_RELEASE_LS:
		list_for_each_entry(op, &ls->ops, list) {
			if (op->n == n && op->type == NLDLM_CMD_NTF_NEW_LS) {
					list_del(&op->list);
					goto out;
			}
		}

		op = calloc(1, sizeof(*op));
		if (!op) {
			nldlmw_ls_event_done_failure(n->nlctx, els->name);
			return;
		}

		op->type = cmd;
		op->n = n;

		list_add_tail(&op->list, &ls->ops);
		break;
	default:
                printf("\"%s\": unknown event\n", els->name);
                break;
	}

out:
	nldlmw_ls_event_done_success(n->nlctx, els->name, ls->ls_global_id);

	ev_timer_stop(loop, &timer);
	ev_timer_set(&timer, COALESCENCE_MEMBER_UPDATE_SECS, 0.);
	ev_timer_start(loop, &timer);
}

static void nl_cb(EV_P_ ev_io *w, int revents)
{
	struct node *n = container_of(w, struct node, w);

	nldlmw_process(n->nlctx);
}

static int netns_switch(const char *nsname)
{
	char net_path[PATH_MAX];
	int netns;

	snprintf(net_path, sizeof(net_path), "%s/%s", NETNS_RUN_DIR, nsname);
	netns = open(net_path, O_RDONLY | O_CLOEXEC);
	if (netns < 0) {
		fprintf(stderr, "Cannot open network namespace \"%s\": %s\n",
			nsname, strerror(errno));
		return -1;
	}

	if (setns(netns, CLONE_NEWNET) < 0) {
		fprintf(stderr, "setting the network namespace \"%s\" failed: %s\n",
			nsname, strerror(errno));
		close(netns);
		return -1;
	}
	close(netns);

	return 0;
}

static int netns_init(struct node *n)
{
	int rv;

	rv = netns_switch(n->nsname);
	if (rv < 0)
		return rv;

	rv = nldlmw_init(&n->nlctx);
	if (rv < 0)
		return rv;

	n->fd = nldlmw_get_fd(n->nlctx);
	printf("File handle %d for %s\n", n->fd, n->nsname);
	return rv;
}

static int dlm_init(const struct node *cur)
{
	struct node *n;
	int rv;

	rv = nldlmw_set_recover_callbacks(cur->nlctx, 1);
	if (rv < 0)
		return rv;

	rv = nldlmw_set_cluster_name(cur->nlctx, "nldlmd");
	if (rv < 0)
		return rv;

	list_for_each_entry(n, &nodes_list, list) {
		rv = nldlmw_new_node(cur->nlctx, n->id, &n->addr);
		if (rv < 0)
			return rv;
	}

	return nldlmw_set_our_nodeid(cur->nlctx, cur->id);
}

int main(int argc, const char *argv[])
{
	struct node *n, *safe;
	int rv;

	if (argc < 2) {
		fprintf(stderr, "config arg failed\n");
		return EXIT_FAILURE;
	}

	rv = config_load(argv[1]);
	if (rv) {
		fprintf(stderr, "config failed\n");
		return EXIT_FAILURE;
	}

	ev_timer_init(&timer, timeout_cb,
		      COALESCENCE_MEMBER_UPDATE_SECS, 0.);
	loop = ev_default_loop(EVBACKEND_EPOLL);

	list_for_each_entry(n, &nodes_list, list) {
		rv = netns_init(n);
		if (rv < 0)
			return -1;

		rv = dlm_init(n);
		if (rv < 0)
			return -1;

		nldlmw_subscribe_events(n->nlctx, "ls-event", print_ls_event, n);

		ev_io_init(&n->w, nl_cb, n->fd, EV_READ);
		ev_io_start(loop, &n->w);
	}

	ev_run(loop, 0);

	list_for_each_entry_safe(n, safe, &nodes_list, list) {
		list_del(&n->list);
		free(n);
	}

	return EXIT_SUCCESS;
}
